@@ -281,7 +281,7 @@ int nvme_resv_release(int fd, __u32 nsid, __u8 rtype, __u8 rrela,
 	__u32 cdw10 = rrela | (iekey ? 1 << 3 : 0) | rtype << 8;
 
 	struct nvme_passthru_cmd cmd = {
-		.opcode		= nvme_cmd_resv_register,
+		.opcode		= nvme_cmd_resv_release,
 		.nsid		= nsid,
 		.cdw10		= cdw10,
 		.addr		= (__u64)(uintptr_t) (payload),
@@ -363,10 +363,8 @@ int nvme_get_log(int fd, __u32 nsid, __u8 log_id, __u32 data_len, void *data)
 		.data_len	= data_len,
 	};
 	__u32 numd = (data_len >> 2) - 1;
-	__le16 numdu, numdl;
+	__u16 numdu = numd >> 16, numdl = numd & 0xffff;
 
-	numdu = htole16((numd & 0xFFFF0000) >> 16);
-	numdl = htole16(numd & 0x0000FFFF);
 	cmd.cdw10 = log_id | (numdl << 16);
 	cmd.cdw11 = numdu;
 
