@@ -35,9 +35,7 @@
 #include <string.h>
 #include <unistd.h>
 #include <math.h>
-#ifdef LIBUDEV_EXISTS
-#include <libudev.h>
-#endif
+#include <dirent.h>
 
 #include <linux/fs.h>
 
@@ -704,29 +702,15 @@ static void get_registers(struct nvme_bar **bar)
 	*bar = membase;
 }
 
+#define MAX_LIST_ITEMS 256
 struct list_item {
 	char                node[1024];
 	struct nvme_id_ctrl ctrl;
 	int                 nsid;
 	struct nvme_id_ns   ns;
 	unsigned            block;
-	__le32              ver;
 };
 
-#ifdef LIBUDEV_EXISTS
-/* For pre NVMe 1.2 devices we must get the version from the BAR, not the
- * ctrl_id.*/
-static void get_version(struct list_item* list_item)
-{
-	struct nvme_bar *bar;
-
-	list_item->ver = list_item->ctrl.ver;
-	if (list_item->ctrl.ver)
-		return;
-	get_registers(&bar);
-	list_item->ver = bar->vs;
-}
-
 static void print_list_item(struct list_item list_item)
 {
 
@@ -748,36 +732,97 @@ static void print_list_item(struct list_item list_item)
 	char format[128];
 	sprintf(format,"%3.0f %2sB + %2d B", (double)lba, l_suffix,
 		list_item.ns.lbaf[(list_item.ns.flbas & 0x0f)].ms);
-	char version[128];
-	sprintf(version,"%d.%d", (list_item.ver >> 16),
-		(list_item.ver >> 8) & 0xff);
 
-	printf("%-16s %-*.*s %-*.*s %-8s %-9d %-26s %-16s %-.*s\n", list_item.node,
+	printf("%-16s %-*.*s %-*.*s %-9d %-26s %-16s %-.*s\n", list_item.node,
             (int)sizeof(list_item.ctrl.sn), (int)sizeof(list_item.ctrl.sn), list_item.ctrl.sn,
             (int)sizeof(list_item.ctrl.mn), (int)sizeof(list_item.ctrl.mn), list_item.ctrl.mn,
-            version, list_item.nsid, usage, format, (int)sizeof(list_item.ctrl.fr), list_item.ctrl.fr);
+            list_item.nsid, usage, format, (int)sizeof(list_item.ctrl.fr), list_item.ctrl.fr);
 }
 
 static void print_list_items(struct list_item *list_items, unsigned len)
 {
 	unsigned i;
 
-	printf("%-16s %-20s %-40s %-8s %-9s %-26s %-16s %-8s\n",
-	    "Node", "SN", "Model", "Version", "Namespace", "Usage", "Format", "FW Rev");
-	printf("%-16s %-20s %-40s %-8s %-9s %-26s %-16s %-8s\n",
+	printf("%-16s %-20s %-40s %-9s %-26s %-16s %-8s\n",
+	    "Node", "SN", "Model", "Namespace", "Usage", "Format", "FW Rev");
+	printf("%-16s %-20s %-40s %-9s %-26s %-16s %-8s\n",
             "----------------", "--------------------", "----------------------------------------",
-            "--------", "---------", "--------------------------", "----------------", "--------");
+            "---------", "--------------------------", "----------------", "--------");
 	for (i = 0 ; i < len ; i++)
 		print_list_item(list_items[i]);
 
 }
-#else
+
+static int get_nvme_info(int fd, struct list_item *item, const char *node)
+{
+	int err;
+
+	err = nvme_identify_ctrl(fd, &item->ctrl);
+	if (err > 0)
+		return err;
+	item->nsid = nvme_get_nsid(fd);
+	err = nvme_identify_ns(fd, item->nsid,
+			       0, &item->ns);
+	if (err > 0)
+		return err;
+	strcpy(item->node, node);
+	item->block = S_ISBLK(nvme_stat.st_mode);
+
+	return 0;
+}
+
+static const char *dev = "/dev/";
+
+/* Assume every block device starting with /dev/nvme is an nvme namespace */
+static int scan_dev_filter(const struct dirent *d)
+{
+	char path[256];
+	struct stat bd;
+
+	if (d->d_name[0] == '.')
+		return 0;
+
+	if (strstr(d->d_name, "nvme")) {
+		snprintf(path, sizeof(path), "%s%s", dev, d->d_name);
+		if (stat(path, &bd))
+			return 0;
+		if (S_ISBLK(bd.st_mode))
+			return 1;
+	}
+	return 0;
+}
+
 static int list(int argc, char **argv, struct command *cmd, struct plugin *plugin)
 {
-	fprintf(stderr,"nvme-list: libudev not detected, install and rebuild.\n");
-	return -1;
+	char path[256];
+	struct dirent **devices;
+	struct list_item *list_items;
+	unsigned int i, n, fd, ret;
+
+	n = scandir(dev, &devices, scan_dev_filter, alphasort);
+	if (n <= 0)
+		return n;
+
+	list_items = calloc(n, sizeof(*list_items));
+	if (!list_items)
+		return ENOMEM;
+
+	for (i = 0; i < n; i++) {
+		snprintf(path, sizeof(path), "%s%s", dev, devices[i]->d_name);
+		fd = open(path, O_RDONLY);
+		ret = get_nvme_info(fd, &list_items[i], path);
+		if (ret)
+			return ret;
+	}
+	print_list_items(list_items, n);
+
+	for (i = 0; i < n; i++)
+		free(devices[i]);
+	free(devices);
+	free(list_items);
+
+	return 0;
 }
-#endif
 
 static int get_nsid()
 {
@@ -792,63 +837,6 @@ static int get_nsid()
 	return nsid;
 }
 
-#ifdef LIBUDEV_EXISTS
-#define MAX_LIST_ITEMS 256
-static int list(int argc, char **argv, struct command *cmd, struct plugin *plugin)
-{
-	struct udev *udev;
-	struct udev_enumerate *enumerate;
-	struct udev_list_entry *devices, *dev_list_entry;
-	struct udev_device *dev;
-
-	struct list_item list_items[MAX_LIST_ITEMS];
-	unsigned count=0;
-
-	udev = udev_new();
-	if (!udev) {
-		perror("nvme-list: Cannot create udev context.");
-		return errno;
-	}
-
-	enumerate = udev_enumerate_new(udev);
-	udev_enumerate_add_match_subsystem(enumerate, "block");
-	udev_enumerate_add_match_property(enumerate, "DEVTYPE", "disk");
-	udev_enumerate_scan_devices(enumerate);
-	devices = udev_enumerate_get_list_entry(enumerate);
-	udev_list_entry_foreach(dev_list_entry, devices) {
-		int err;
-		const char *path, *node;
-		path = udev_list_entry_get_name(dev_list_entry);
-		dev  = udev_device_new_from_syspath(udev, path);
-		node = udev_device_get_devnode(dev);
-		if (strstr(node,"nvme")!=NULL){
-			open_dev(node);
-			err = nvme_identify_ctrl(fd, &list_items[count].ctrl);
-			if (err > 0)
-				return err;
-			list_items[count].nsid = nvme_get_nsid(fd);
-			err = nvme_identify_ns(fd, list_items[count].nsid,
-						0, &list_items[count].ns);
-			if (err > 0)
-				return err;
-			strcpy(list_items[count].node, node);
-			list_items[count].block = S_ISBLK(nvme_stat.st_mode);
-			get_version(&list_items[count]);
-			count++;
-		}
-	}
-	udev_enumerate_unref(enumerate);
-	udev_unref(udev);
-
-	if (count)
-		print_list_items(list_items, count);
-	else
-		printf("No NVMe devices detected.\n");
-
-	return 0;
-}
-#endif
-
 int __id_ctrl(int argc, char **argv, struct command *cmd, struct plugin *plugin, void (*vs)(__u8 *vs))
 {
 	const char *desc = "Send an Identify Controller command to "\
@@ -1306,34 +1294,92 @@ static void print_lo_hi_64(uint32_t *val)
 
 static int show_registers(int argc, char **argv, struct command *cmd, struct plugin *plugin)
 {
-	int opt, long_index;
+	const char *desc = "Reads and shows the defined NVMe controller registers "\
+					"in binary or human-readable format";
+	const char *human_readable = "show info in readable format";
 	struct nvme_bar *bar;
 
-	while ((opt = getopt_long(argc, (char **)argv, "", NULL,
-					&long_index)) != -1);
+	struct config {
+		int human_readable;
+	};
 
-	check_arg_dev(argc, argv);
-	devicename = basename(argv[optind]);
+	struct config cfg = {
+		.human_readable = 0,
+	};
+
+	const struct argconfig_commandline_options command_line_options[] = {
+		{"human-readable",  'H', "", CFG_NONE, &cfg.human_readable,  no_argument, human_readable},
+		{0}
+	};
+
+	parse_and_open(argc, argv, desc, command_line_options, &cfg, sizeof(cfg));
 
 	get_registers(&bar);
-	printf("cap     : ");
-	print_lo_hi_64((uint32_t *)&bar->cap);
 
-	printf("version : %x\n", bar->vs);
-	printf("intms   : %x\n", bar->intms);
-	printf("intmc   : %x\n", bar->intmc);
-	printf("cc      : %x\n", bar->cc);
-	printf("csts    : %x\n", bar->csts);
-	printf("nssr    : %x\n", bar->nssr);
-	printf("aqa     : %x\n", bar->aqa);
-	printf("asq     : ");
-	print_lo_hi_64((uint32_t *)&bar->asq);
+	if (cfg.human_readable) {
+		printf("cap     : ");
+		print_lo_hi_64((uint32_t *)&bar->cap);
+		show_registers_cap(&bar->cap);
 
-	printf("acq     : ");
-	print_lo_hi_64((uint32_t *)&bar->acq);
+		printf("version : %x\n", bar->vs);
+		show_registers_version(bar->vs);
 
-	printf("cmbloc  : %x\n", bar->cmbloc);
-	printf("cmbsz   : %x\n", bar->cmbsz);
+		printf("intms   : %x\n", bar->intms);
+		printf("\tInterrupt Vector Mask Set (IVMS): %x\n\n", bar->intms); 
+
+		printf("intmc   : %x\n", bar->intmc);
+		printf("\tInterrupt Vector Mask Clear (IVMC): %x\n\n", bar->intmc); 
+
+		printf("cc      : %x\n", bar->cc);
+		show_registers_cc(bar->cc);
+
+		printf("csts    : %x\n", bar->csts);
+		show_registers_csts(bar->csts);
+
+		printf("nssr    : %x\n", bar->nssr);
+		printf("\tNVM Subsystem Reset Control (NSSRC): %u\n\n", bar->nssr); 
+
+		printf("aqa     : %x\n", bar->aqa);
+		show_registers_aqa(bar->aqa);
+
+		printf("asq     : ");
+		print_lo_hi_64((uint32_t *)&bar->asq);
+		printf("\tAdmin Submission Queue Base (ASQB): ");
+		print_lo_hi_64((uint32_t *)&bar->asq);
+		printf("\n");
+
+		printf("acq     : ");
+		print_lo_hi_64((uint32_t *)&bar->acq);
+		printf("\tAdmin Completion Queue Base (ACQB): ");
+		print_lo_hi_64((uint32_t *)&bar->acq);
+		printf("\n");
+
+		printf("cmbloc  : %x\n", bar->cmbloc);
+		show_registers_cmbloc(bar->cmbloc, bar->cmbsz);
+
+		printf("cmbsz   : %x\n", bar->cmbsz);
+		show_registers_cmbsz(bar->cmbsz);
+	}
+	else {
+		printf("cap     : ");
+		print_lo_hi_64((uint32_t *)&bar->cap);
+
+		printf("version : %x\n", bar->vs);
+		printf("intms   : %x\n", bar->intms);
+		printf("intmc   : %x\n", bar->intmc);
+		printf("cc      : %x\n", bar->cc);
+		printf("csts    : %x\n", bar->csts);
+		printf("nssr    : %x\n", bar->nssr);
+		printf("aqa     : %x\n", bar->aqa);
+		printf("asq     : ");
+		print_lo_hi_64((uint32_t *)&bar->asq);
+
+		printf("acq     : ");
+		print_lo_hi_64((uint32_t *)&bar->acq);
+
+		printf("cmbloc  : %x\n", bar->cmbloc);
+		printf("cmbsz   : %x\n", bar->cmbsz);
+	}
 
 	return 0;
 }
@@ -1639,7 +1685,8 @@ static int write_uncor(int argc, char **argv, struct command *cmd, struct plugin
 	if (err < 0)
 		return errno;
 	else if (err != 0)
-		fprintf(stderr, "NVME Write Uncorrectable Command Error:%d\n", err);
+		fprintf(stderr, "NVME IO command error:%s(%x)\n",
+					nvme_status_to_string(err), err);
 	else
 		printf("NVME Write Uncorrectable Success\n");
 	return err;
@@ -1713,7 +1760,8 @@ static int write_zeroes(int argc, char **argv, struct command *cmd, struct plugi
 	if (err < 0)
 		return errno;
 	else if (err != 0)
-		fprintf(stderr, "NVME Write Zeroes Command Error:%d\n", err);
+		fprintf(stderr, "NVME IO command error:%s(%x)\n",
+					nvme_status_to_string(err), err);
 	else
 		printf("NVME Write Zeroes Success\n");
 	return err;
@@ -2111,6 +2159,8 @@ static int submit_io(int opcode, char *command, const char *desc,
 	int flags = opcode & 1 ? O_RDONLY : O_WRONLY | O_CREAT;
 	int mode = S_IRUSR | S_IWUSR |S_IRGRP | S_IWGRP| S_IROTH;
 	__u16 control = 0;
+	int phys_sector_size = 0;
+	long long buffer_size = 0;
 
 	const char *start_block = "64-bit addr of first block to access";
 	const char *block_count = "number of blocks (zeroes based) on device to access";
@@ -2209,7 +2259,18 @@ static int submit_io(int opcode, char *command, const char *desc,
 		return EINVAL;
 	}
 
-	if (posix_memalign(&buffer, getpagesize(), cfg.data_size))
+	if (ioctl(fd, BLKPBSZGET, &phys_sector_size) < 0)
+		return errno;
+
+	buffer_size = (cfg.block_count + 1) * phys_sector_size;
+	if (cfg.data_size < buffer_size) {
+		fprintf(stderr, "Rounding data size to fit block count (%lld bytes)\n",
+				buffer_size);
+	} else {
+		buffer_size = cfg.data_size;
+	}
+
+	if (posix_memalign(&buffer, getpagesize(), buffer_size))
 		return ENOMEM;
 	memset(buffer, 0, cfg.data_size);
 
@@ -2622,7 +2683,7 @@ int main(int argc, char **argv)
 	setlocale(LC_ALL, "");
 
 	ret = handle_plugin(argc - 1, &argv[1], nvme.extensions);
-	if (ret == -1)
+	if (ret == -ENOTTY)
 		general_help(&builtin);
 
 	return ret;
